#ifndef CHILD_PROCESS_H
#define CHILD_PROCESS_H

/*
 * OO base class for process functions and child processes
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <sys/types.h>
#include <tagcoll/Exception.h>

class Process
{
protected:
	void detachFromTTY() throw (Tagcoll::SystemException);

public:
	virtual ~Process() {}

	// TODO: since the destructor is called twice (one in the parent and one in
	// the child), it could be useful to add a bool isChild() method to let the
	// destructor and other functions know where they are operating.  The value
	// returned can be set by ChildProcess::fork.
	virtual int main() throw () = 0;

	/// Pretty-print the return value of a process into a string
	static std::string formatStatus(int status) throw ();
	
	/// Change working directory
	static void chdir(const std::string& dir) throw (Tagcoll::SystemException);

	/// Change root directory
	static void chroot(const std::string& dir) throw (Tagcoll::SystemException);

	/// Change umask (always succeeds and returns the previous umask)
	static mode_t umask(mode_t mask) throw ();

	/// Set user and group permissions
	static void setPerms(const std::string& user)
		throw (Tagcoll::ConsistencyCheckException, Tagcoll::SystemException);
	static void setPerms(const std::string& user, const std::string& group)
		throw (Tagcoll::ConsistencyCheckException, Tagcoll::SystemException);
	static void setPerms(uid_t user)
		throw (Tagcoll::ConsistencyCheckException, Tagcoll::SystemException);
	static void setPerms(uid_t user, gid_t group)
		throw (Tagcoll::ConsistencyCheckException, Tagcoll::SystemException);

	/// Get current resource limits; store also maximum resource limits in max
	/// if nonzero
	static int getCPUTimeLimit(int* max = 0) throw (Tagcoll::SystemException);
	static int getFileSizeLimit(int* max = 0) throw (Tagcoll::SystemException);
	static int getDataMemoryLimit(int* max = 0) throw (Tagcoll::SystemException);
	static int getChildrenLimit(int* max = 0) throw (Tagcoll::SystemException);
	static int getOpenFilesLimit(int* max = 0) throw (Tagcoll::SystemException);
	static int getCoreSizeLimit(int* max = 0) throw (Tagcoll::SystemException);

	/// Set resource limits
	static void setCPUTimeLimit(int value) throw (Tagcoll::SystemException);
	static void setFileSizeLimit(int value) throw (Tagcoll::SystemException);
	static void setDataMemoryLimit(int value) throw (Tagcoll::SystemException);
	static void setChildrenLimit(int value) throw (Tagcoll::SystemException);
	static void setOpenFilesLimit(int value) throw (Tagcoll::SystemException);
	static void setCoreSizeLimit(int value) throw (Tagcoll::SystemException);
};

struct rusage;

class ChildProcess
{
protected:
	Process* _proc;
	pid_t _pid;
	
public:
	ChildProcess(Process* proc) throw () : _proc(proc), _pid(-1) {}
	~ChildProcess() throw () {}

	/// For a subprocess to run proc.
	/// After this function exits successfully, `proc' can safely be deleted.
	pid_t ChildProcess::fork() throw (Tagcoll::SystemException);

	/// Fork a subprocess to run proc.  If one of the std*fd variables is
	/// non-null, create a pipe connected to the corresponding file descriptor
	/// of the child process and store the parent end in the std*fd variable.
	/// After this function exits successfully, `proc' can safely be deleted.
	pid_t forkAndRedirect(int* stdinfd = 0, int* stdoutfd = 0, int* stderrfd = 0) throw (Tagcoll::SystemException);
	
	/// Get the pid of the child process or (pid_t)-1 if no child is running
	pid_t pid() const throw () { return _pid; }

	/// Wait for the child to finish, returing its exit status.  Return -1 if
	/// no child is running.
	/// TODO: gracefully handle the EINTR error code
	int wait() throw (Tagcoll::SystemException, Tagcoll::InterruptedException);

	/// Wait for the child to finish, returing its exit status and storing
	/// resource usage informations in `ru'.  Return -1 if no child is running.
	/// TODO: gracefully handle the EINTR error code
	int wait(struct rusage* ru) throw (Tagcoll::SystemException, Tagcoll::InterruptedException);

	/// Send the given signal to the process
	void kill(int signal) throw (Tagcoll::SystemException);
};

// vim:set ts=4 sw=4:
#endif
