/*
 * debtags - Implement package tags support for Debian
 *
 * Copyright (C) 2003,2004,2005  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef PRINTER_H
#define PRINTER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#define APPNAME PACKAGE_NAME
#else
#warning No config.h found: using fallback values
#define APPNAME __FILE__
#define PACKAGE_VERSION "unknown"
#endif

#include <apt-front/cache/entity/tag.h>
#include <apt-front/cache/entity/package.h>
#include <tagcoll/TextFormat.h>
#include <tagcoll/ItemGrouper.h>

#include <errno.h>


template<typename ITEM, typename TAG>
class Printer : public Tagcoll::Consumer<ITEM, TAG>
{
public:
	virtual void flush() {}
};

class APTPrinter : public Printer<aptFront::cache::entity::Package, aptFront::cache::entity::Tag>
{
protected:
	typedef aptFront::cache::entity::Package Package;
	typedef aptFront::cache::entity::Tag Tag;

	int _count;
	Tagcoll::OpSet<Package> blacklist;

	virtual void consumeItemUntagged(const Package& pkg)
	{
		if (blacklist.find(pkg) == blacklist.end())
		{
			printf("%.*s - %.*s\n", PFSTR(pkg.name()), PFSTR(pkg.shortDescription("(short description not available)")));
			_count++;
		}
	}
	virtual void consumeItem(const Package& pkg, const Tagcoll::OpSet<Tag>& tags)
	{
		consume(pkg);
	}

public:
	APTPrinter() : _count(0) {}
	APTPrinter(const Tagcoll::OpSet<Package>& blacklist)
		: _count(0), blacklist(blacklist) {}
	virtual ~APTPrinter() {}

	int count() const { return _count; }

	void print(const Tagcoll::OpSet<Package>& pkgs)
	{
		consumeItemsUntagged(pkgs);
	}
};

class NullPrinter : public Printer<aptFront::cache::entity::Package, aptFront::cache::entity::Tag>
{
protected:
	typedef aptFront::cache::entity::Package Package;
	typedef aptFront::cache::entity::Tag Tag;

	virtual void consumeItemUntagged(const Package& pkg)
	{
	}
	virtual void consumeItem(const Package& pkg, const Tagcoll::OpSet<Tag>& tags)
	{
	}
	virtual void consumeItemsUntagged(const Tagcoll::OpSet<Package>& pkgs)
	{
	}
	virtual void consumeItems(const Tagcoll::OpSet<Package>& pkgs, const Tagcoll::OpSet<Tag>& tags)
	{
	}

public:
	NullPrinter() {}
};

class TagcollPrinter : public Printer<aptFront::cache::entity::Package, aptFront::cache::entity::Tag>
{
protected:
	typedef aptFront::cache::entity::Package Package;
	typedef aptFront::cache::entity::Tag Tag;

	Tagcoll::Converter<Package, std::string> fromitem;
	Tagcoll::Converter<Tag, std::string> fromtag;
	Tagcoll::TextFormat<Package, Tag> writer;

	virtual void consumeItemUntagged(const Package& pkg)
	{
		writer.consume(pkg);
	}
	virtual void consumeItem(const Package& pkg, const Tagcoll::OpSet<Tag>& tags)
	{
		writer.consume(pkg, tags);
	}

public:
	TagcollPrinter() : writer(fromitem, fromtag, stdout) {}
};

class GroupedTagcollPrinter : public Printer<aptFront::cache::entity::Package, aptFront::cache::entity::Tag>
{
protected:
	typedef aptFront::cache::entity::Package Package;
	typedef aptFront::cache::entity::Tag Tag;

	Tagcoll::Converter<Package, string> fromitem;
	Tagcoll::Converter<Tag, string> fromtag;
	Tagcoll::TextFormat<Package, Tag> writer;
	Tagcoll::ItemGrouper<Package, Tag> grouper;

	virtual void consumeItemUntagged(const Package& pkg)
	{
		grouper.consume(pkg);
	}
	virtual void consumeItem(const Package& pkg, const Tagcoll::OpSet<Tag>& tags)
	{
		grouper.consume(pkg, tags);
	}
	virtual void consumeItemsUntagged(const Tagcoll::OpSet<Package>& pkgs)
	{
		grouper.consume(pkgs);
	}
	virtual void consumeItems(const Tagcoll::OpSet<Package>& pkgs, const Tagcoll::OpSet<Tag>& tags)
	{
		grouper.consume(pkgs, tags);
	}

public:
	GroupedTagcollPrinter() : writer(fromitem, fromtag, stdout) {}

	virtual void flush()
	{
		grouper.output(writer);
	}
};

class Installer : public Printer<aptFront::cache::entity::Package, aptFront::cache::entity::Tag>
{
protected:
	typedef aptFront::cache::entity::Package Package;
	typedef aptFront::cache::entity::Tag Tag;

	Tagcoll::OpSet<Package> selection;

	virtual void consumeItemUntagged(const Package& pkg)
	{
		selection += pkg;
	}
	virtual void consumeItem(const Package& pkg, const Tagcoll::OpSet<Tag>& tags)
	{
		selection += pkg;
	}
	virtual void consumeItemsUntagged(const Tagcoll::OpSet<Package>& pkgs)
	{
		selection += pkgs;
	}
	virtual void consumeItems(const Tagcoll::OpSet<Package>& pkgs, const Tagcoll::OpSet<Tag>& tags)
	{
		selection += pkgs;
	}

public:
	virtual void flush()
	{
		const char* parms[selection.size() + 1];
		int i = 0;
		parms[i++] = APTGET;
		parms[i++] = "install";
		for (Tagcoll::OpSet<Package>::iterator j = selection.begin();
				j != selection.end(); j++)
			parms[i++] = j->name().c_str();
		parms[i] = 0;
		execv(APTGET, (char* const*)parms);
		throw Tagcoll::SystemException(errno, "Running apt-get");
	}
};

// vim:set ts=4 sw=4:
#endif
