/*
 * debtags - Implement package tags support for Debian
 *
 * Copyright (C) 2003,2004,2005  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#define APPNAME PACKAGE_NAME
#else
#warning No config.h found: using fallback values
#define APPNAME __FILE__
#define PACKAGE_VERSION "unknown"
#endif

#include <apt-front/init.h>
#include <apt-front/cache/cache.h>
#include <apt-front/cache/component/packagetags.h>
#include <apt-front/cache/component/tags.h>
#include <apt-front/cache/component/debtags/expression.h>
#include <apt-front/cache/component/debtags/update.h>
#include <apt-front/cache/component/debtags/utils.h>
#include <apt-front/cache/entity/tag.h>
#include <apt-front/predicate/predicate.h>
#include <apt-front/predicate/factory.h>
#include <apt-front/utils/paths.h>

#include <tagcoll/StdioParserInput.h>
#include <tagcoll/TextFormat.h>
#include <tagcoll/SmartHierarchy.h>
#include <tagcoll/InputMerger.h>
#include <tagcoll/experiments.h>

#include "Environment.h"
#include "CommandlineParser.h"
#include "acqprogress.h"
#include "Printer.h"

#include <apt-pkg/configuration.h>

#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/types.h>	// umask
#include <sys/stat.h>	// umask

using namespace std;
using namespace aptFront;
using namespace aptFront::cache;
using namespace aptFront::cache::entity;
using namespace aptFront::cache::component;
using namespace aptFront::cache::component::debtags;
using namespace Tagcoll;

template<typename TAG, typename _Traits>
basic_ostream<char, _Traits>& operator<<(basic_ostream<char, _Traits>& out, const Tagcoll::OpSet<TAG>& tags)
{
	for (typename Tagcoll::OpSet<TAG>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i == tags.begin())
			out << i->fullname();
		else
			out << ", " << i->fullname();
	return out;
}


component::PackageTags& debtagsInit(bool write = false)
{
	    aptFront::init();

		// Open the package cache
		if (write)
			cache::Global::get().open(
					  Cache::OpenDefault
					| Cache::OpenReadOnly
					| Cache::OpenTags
					| Cache::OpenDebtags
					| Cache::OpenDebtagsEditable);
		else
			cache::Global::get().open(
					  Cache::OpenDefault
					| Cache::OpenReadOnly
					| Cache::OpenTags
					| Cache::OpenDebtags);

		return Global::get().debtags();
}

void wantTagDatabase() throw ()
{
	if (!component::PackageTags::hasTagDatabase())
		fatal_error(
				"The tag database has not yet been generated: you need to run \"debtags update\" (as root) before using the package tags\n");
}

entity::Package packageByName(const std::string& name)
{
	return Global::get().packages().packageByName(name);
}

void readCollection(const string& file, Tagcoll::Consumer<string, string>& output)
	throw (FileException, ParserException)
{
	Tagcoll::Converter<string, string> conv;
	if (file == "-")
	{
		StdioParserInput input(stdin, "<stdin>");
		TextFormat<string, string>::parse(conv, conv, input, output);
	}
	else
	{
		StdioParserInput input(file);
		TextFormat<string, string>::parse(conv, conv, input, output);
	}
}

template<class ITEM, class TAG>
void readCollection(
		Tagcoll::Converter<string, ITEM>& toitem,
		Tagcoll::Converter<string, TAG>& totag,
		const string& file,
		Tagcoll::Consumer<ITEM, TAG>& output)
	throw (FileException, ParserException)
{
	if (file == "-")
	{
		StdioParserInput input(stdin, "<stdin>");
		TextFormat<ITEM, TAG>::parse(toitem, totag, input, output);
	}
	else
	{
		StdioParserInput input(file);
		TextFormat<ITEM, TAG>::parse(toitem, totag, input, output);
	}
}

class SubstringTagMatcher
{
protected:
	vector<string> patterns;

public:
	void add(const std::string& pattern) { patterns.push_back(pattern); }
	
	bool operator()(const Facet& item)
	{
		for (vector<string>::const_iterator i = patterns.begin();
				i != patterns.end(); i++)
		{
			if (strcasestr(item.name().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.shortDescription().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.longDescription().c_str(), i->c_str()))
				return true;
		}
		return false;
	}
	bool operator()(const Tag& item)
	{
		//cerr << "Testing: " << item.fullname() << endl;
		for (vector<string>::const_iterator i = patterns.begin();
				i != patterns.end(); i++)
		{
			if (strcasestr(item.fullname().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.shortDescription().c_str(), i->c_str()))
				return true;
			if (strcasestr(item.longDescription().c_str(), i->c_str()))
				return true;
		}
		//cerr << "  not passed." << endl;
		return false;
	}
};

// Catch the tags for a given item
/*
class TagcollCatcher : public Tagcoll::Consumer<string, string>
{
protected:
	string _item;
	OpSet<string> _tags;
	
public:
	TagcollCatcher(const std::string& item) throw ()
		: _item(item) {}
	virtual ~TagcollCatcher() throw () {}

	const OpSet<string>& tags() const throw () { return _tags; }

	// If the package is really untagged, there's nothing we can do here
	virtual void consume(const std::string& item) throw () {}
	virtual void consume(const OpSet<string>& items) throw () {}
	
	// Catch the tags that belong to _item
	virtual void consume(const std::string& item, const OpSet<string>& tags) throw ()
	{
		if (item == _item)
			_tags += tags;
	}

	// Catch the tags that belong to _item
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
	{
		if (items.contains(_item))
			_tags += tags;
	}
};
*/

// Check a collection agaisnt a tag vocabulary
class TagcollChecker : public Tagcoll::Consumer<string, string>
{
protected:
	const component::Tags& voc;
	int missing_count;
	map<string, int> missing;
	unsigned int mtag_width;
	int max_missing;

	// If the package is really untagged, there's nothing we can do here
	virtual void consumeItemUntagged(const std::string& item) {}
	virtual void consumeItemsUntagged(const OpSet<string>& items) {}
	
	// Catch the tags that belong to _item
	virtual void consumeItem(const std::string& item, const OpSet<string>& tags)
	{
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			checkTag(*i);
	}

	// Catch the tags that belong to _item
	virtual void consumeItems(const OpSet<string>& items, const OpSet<string>& tags)
	{
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			checkTag(*i);
	}

	void checkTag(const std::string& tag)
	{
		if (!voc.hasTag(tag))
		{
			//warning("Tag %.*s not found in vocabulary\n", PFSTR(tag));
			missing_count++;
			missing[tag]++;
			if (tag.size() > mtag_width)
				mtag_width = tag.size();
			if (missing[tag] > max_missing)
				max_missing = missing[tag];
		}
	}
	
public:
	TagcollChecker()
		: voc(Global::get().tags()), missing_count(0), mtag_width(0), max_missing(0) {}
	virtual ~TagcollChecker() {}

	// Get the number of times a bad tag has been found in input
	int missingCount() const { return missing_count; }

	void report()
	{
		if (missing_count > 0)
		{
			warning("%d tags were found in packages but not in the vocabulary\n", missing.size());
			warning("This happened %d times\n", missing_count);
			warning("The tags found in the collection but not in the vocabulary are:\n");

			int mtag_maxdigits = 0;
			for (int i = max_missing; i > 0; i = i / 10)
				mtag_maxdigits++;

			for (map<string, int>::const_iterator i = missing.begin();
					i != missing.end(); i++)
			{
				warning("\t%.*s %*s (in %*d package%s)\n",
					PFSTR(i->first), 
					mtag_width - i->first.size(),
					"", 
					mtag_maxdigits,
					i->second,
					i->second > 1 ? "s" : "");
			}
		}
	}
};

// Score tags by how many installed packages are there with them
class TagsScorer : public map<entity::Tag, int>, public Tagcoll::Consumer<entity::Package, entity::Tag>
{
protected:
	Expression ignore;

	virtual void consumeItemUntagged( const entity::Package& ) {}
	virtual void consumeItemsUntagged( const OpSet<entity::Package>& ) {}

	virtual void consumeItem(const entity::Package& item, const OpSet<entity::Tag>& tags)
	{
		if (item.isInstalled() || item.isUpgradable() || item.isBroken())
			for (OpSet<Tag>::const_iterator i = tags.begin(); i != tags.end(); i++)
			{
				OpSet<Tag> it;
				it += *i;
				if (!ignore(it))
				{
					iterator j = find(*i);
					if (j == end())
						insert(make_pair<Tag, int>(*i, 1));
					else
						j->second++;
				}
			}
	}

public:
	TagsScorer() : ignore("special::not-yet-tagged*") {}
	virtual ~TagsScorer() {}

	int score(const entity::Tag& tag) const
	{
		const_iterator i = find(tag);
		if (i == end())
			return 0;
		else
			return i->second;
	}
};

// Needs package for TagsScorer
class PackageScorer : public map<int, entity::Package>, public Tagcoll::Consumer<entity::Package, Tag>
{
protected:
	const TagsScorer& scores;

	virtual void consumeItemUntagged(const entity::Package& item) {}
	virtual void consumeItemsUntagged(const OpSet<entity::Package>& items) {}

	virtual void consumeItem(const entity::Package& item, const OpSet<Tag>& tags) throw ()
	{
		if (item.isInstalled() || item.isUpgradable() || item.isBroken())
		{
			int score = 0;
			for (OpSet<Tag>::const_iterator i = tags.begin();
					i != tags.end(); i++)
				score += scores.score(*i);

			insert(make_pair<int, entity::Package>(score, item));
		}
	}

public:
	PackageScorer(const TagsScorer& scores) throw () : scores(scores) {}
	virtual ~PackageScorer() throw () {}

};

template<class ITEM>
class FacetcollPrinter : public Tagcoll::Consumer<ITEM, Facet>
{
protected:
	virtual void consumeItemUntagged(const ITEM& item)
	{
		printf("%.*s\n", PFSTR(item));
	}
	virtual void consumeItem(const ITEM& item, const OpSet<Facet>& tags)
	{
		printf("%.*s: ", PFSTR(item));
		for (OpSet<Facet>::const_iterator i = tags.begin();
				i != tags.end(); i++)
			if (i == tags.begin())
				printf("%.*s", PFSTR(i->name()));
			else
				printf(", %.*s", PFSTR(i->name()));
		printf("\n");

	}

public:
	virtual ~FacetcollPrinter() {}
};

template<>
void FacetcollPrinter<entity::Package>::consumeItemUntagged(const entity::Package& item)
{
	printf("%.*s\n", PFSTR(item.name()));
}
template<>
void FacetcollPrinter<entity::Package>::consumeItem(const entity::Package& item, const OpSet<Facet>& tags)
{
	printf("%.*s: ", PFSTR(item.name()));
	for (OpSet<Facet>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i == tags.begin())
			printf("%.*s", PFSTR(i->name()));
		else
			printf(", %.*s", PFSTR(i->name()));
	printf("\n");

}

template< typename PKG, typename TAG >
class TagcollConsumerAdaptor : public utils::ConsumerImpl< PKG, TagcollConsumerAdaptor<PKG, TAG> >
{
protected:
	Consumer<PKG, TAG>& m_out;

public:
	TagcollConsumerAdaptor( Consumer<PKG, TAG>& out) : m_out( out ) {}
	virtual void consume( const PKG& a ) {
		if (a != PKG())
			m_out.consume(a, a.tags());
	}
};

template< typename ITEM >
class RangeCounter : public utils::ConsumerImpl< ITEM, RangeCounter<ITEM> >
{
protected:
	int* count;

public:
	RangeCounter(int* count) : count(count) {}
	virtual void consume(const ITEM& a) {
		++*count;
	}
	
};

class Searcher
{
protected:
	component::PackageTags& debtags;
	Printer<Package, Tag>* printer;

public:
	Searcher(component::PackageTags& debtags, Printer<Package, Tag>* printer) :
		debtags(debtags), printer(printer) {}

	void output()
	{
		debtags.outputPatched(*printer);
		printer->flush();
	}

	int output(const string& expression, bool invertMatch)
	{
		// Build the grep filter chain
		FilterItemsByExpression<Package, Tag> filter(*printer, expression);
		if (invertMatch)
			filter.setMatchType(FilterItemsByExpression<Package, Tag>::INVERTED);
		debtags.outputPatched(filter);
		printer->flush();
		return filter.countMatched();
	}

	int output(predicate::Predicate<Package> p)
	{
		aptFront::cache::component::Packages& packages = Global::get().packages();
		utils::Range<Package> r = filteredRange(
				utils::Range<Package>(range(packages.packagesBegin(), packages.packagesEnd())), p);
		r.output(TagcollConsumerAdaptor<Package, Tag>(*printer));
		printer->flush();
		int count = 0;
		RangeCounter<Package> counter(&count);
		r.output(counter);
		return count;
	}
};

class SmartSearcher : public Printer<Package, Tag>, private CardinalityStore<Package, Tag>
{
protected:
	component::PackageTags& debtags;
	Printer<Package, Tag>* printer;
	map<Tag, size_t> tagCount;
	OpSet<Tag> m_topTags;

	virtual void consumeItemUntagged(const Package& pkg)
	{
		printer->consume(pkg);
		CardinalityStore<Package, Tag>::consumeItemUntagged(pkg);
	}
	virtual void consumeItem(const Package& pkg, const Tagcoll::OpSet<Tag>& tags)
	{
		printer->consume(pkg, tags);
		CardinalityStore<Package, Tag>::consumeItem(pkg, tags);
		for (OpSet<Tag>::const_iterator i = tags.begin();
				i != tags.end(); i++)
			++tagCount[*i];
	}
	virtual void consumeItemsUntagged(const Tagcoll::OpSet<Package>& pkgs)
	{
		printer->consume(pkgs);
		CardinalityStore<Package, Tag>::consumeItemsUntagged(pkgs);
	}
	virtual void consumeItems(const Tagcoll::OpSet<Package>& pkgs, const Tagcoll::OpSet<Tag>& tags)
	{
		printer->consume(pkgs, tags);
		CardinalityStore<Package, Tag>::consumeItems(pkgs, tags);
		for (OpSet<Tag>::const_iterator i = tags.begin();
				i != tags.end(); i++)
			tagCount[*i] += pkgs.size();
	}

	vector< pair<Tag, size_t> > getTopTags(size_t count)
	{
		vector< pair<Tag, size_t> > res;
		for (map<Tag, size_t>::const_iterator i = tagCount.begin();
				i != tagCount.end(); i++)
		{
			pair<Tag, size_t> hand = *i;
			for (size_t j = 0; j < count; j++)
			{
				if (j >= res.size())
				{
					res.push_back(hand);
					break;
				}
				else if (hand.second > res[j].second)
				{
					pair<Tag, size_t> tmp = res[j];
					res[j] = hand;
					hand = tmp;
				}
			}
		}
		return res;
	}

public:
	SmartSearcher(component::PackageTags& debtags, Printer<Package, Tag>* printer) :
		debtags(debtags), printer(printer) {}

	void clear()
	{
		CardinalityStore<Package, Tag>::operator=(CardinalityStore<Package, Tag>());
		tagCount.clear();
		m_topTags.clear();
	}

	OpSet<Tag> topTags() { return m_topTags; }

	int outputRelated()
	{
		const size_t relevantTags = 5;
		vector< pair<Tag, size_t> > topTagList = getTopTags(relevantTags);
		Scores<Tag> scores;
		vector< OpSet<Tag> > tagsets;
		float distance = relevantTags;
		int count = 0;

		m_topTags.clear();
		for (vector< pair<Tag, size_t> >::const_iterator i = topTagList.begin();
				i != topTagList.end(); i++)
			m_topTags += i->first;

		// Get the tagsets that score less in weighted distance from topTags
		for (tagsets_t::const_iterator i = this->tagsets.begin();
				i != this->tagsets.end(); i++)
		{
			float d = scores.distance(m_topTags, i->first);
			if (d <= distance)
			{
				if (d < distance)
				{
					tagsets.clear();
					distance = d;
				}
				tagsets.push_back(i->first);
			}
		}

		// print the items they have
		for (vector< OpSet<Tag> >::const_iterator i = tagsets.begin();
				i != tagsets.end(); i++)
		{
			OpSet<Package> pkgs = debtags.tagdb().getItems(*i);
			printer->consume(pkgs, *i);
			count += pkgs.size();
		}
		return count;
	}

	virtual void flush()
	{
		printer->flush();
	}
};


/*
class ItemPrinter : public Tagcoll::Tagcoll::Consumer<string, string>
{
public:
	virtual ~ItemPrinter() throw () {}

	virtual void consume(const std::string& item) throw ()
	{
		printf("%.*s\n", PFSTR(item));
	}
	virtual void consume(const std::string& item, const OpSet<std::string>& tags) throw ()
	{
		printf("%.*s\n", PFSTR(item));
	}
};
*/

// Needs Package to see if it's installed
class TODOFilter : public Tagcoll::FilterItemsByExpression<entity::Package, Tag>
{
protected:
	void consumeItemUntagged(const entity::Package& pkg)
	{
		if (pkg.isInstalled() || pkg.isUpgradable() || pkg.isBroken())
			Tagcoll::FilterItemsByExpression<entity::Package, Tag>::consumeItemUntagged(pkg);
	}

	void consumeItem(const entity::Package& pkg, const OpSet<Tag>& tags)
	{
		if (pkg.isInstalled() || pkg.isUpgradable() || pkg.isBroken())
			Tagcoll::FilterItemsByExpression<entity::Package, Tag>::consumeItem(pkg, tags);
	}

public:
	TODOFilter()
		: Tagcoll::FilterItemsByExpression<entity::Package, Tag>("special::not-yet-tagged*") {}
	TODOFilter(Tagcoll::Consumer<entity::Package, Tag>&cons)
		: Tagcoll::FilterItemsByExpression<entity::Package, Tag>(cons, "special::not-yet-tagged*") {}
};


template<class ITEM>
class ExprCollector : public Tagcoll::Consumer<ITEM, Tag>, public OpSet<entity::Package>
{
protected:
	Expression expr;

	virtual void consumeItemUntagged(const entity::Package& item) throw ()
	{
		if (expr(OpSet<Tag>()))
			(*this) += item;
	}
	virtual void consumeItem(const entity::Package& item, const OpSet<Tag>& tags) throw ()
	{
		if (expr(tags))
			(*this) += item;
	}

public:
	ExprCollector(const std::string& expression)
		: expr(expression) {}
	virtual ~ExprCollector() {}
};

// Might need package in the future to count stats about installed packages
// TODO: count stats about installed packages
class StatsCollector : public Tagcoll::Consumer<entity::Package, Tag>
{
protected:
	Expression has_nyt;
	Expression is_ct;
	int stat_seen;
	int stat_onlynyt;
	int stat_nyt;
	int stat_notags;
	int stat_tagged;
	int stat_complete;

	virtual void consumeItemUntagged(const entity::Package& item) throw ()
	{
		stat_notags++;
		stat_seen++;
	}
	virtual void consumeItem(const entity::Package& item, const OpSet<Tag>& tags) throw ()
	{
		if (tags.empty()) {
			stat_notags++;
		} else if (has_nyt(tags)) {
			stat_nyt++;
			stat_onlynyt++;
			for (OpSet<Tag>::const_iterator i = tags.begin();
					i != tags.end(); i++)
				if (i->fullname().substr(0, 23) != "special::not-yet-tagged")
				{
					stat_onlynyt--;
					break;
				}
		} else {
			stat_tagged++;
			if (is_ct(tags))
				stat_complete++;
		}
		stat_seen++;
	}

public:
	StatsCollector()
		: has_nyt("special::not-yet-tagged*"),
		  is_ct("special::completely-tagged"),
		  stat_seen(0), stat_onlynyt(0), stat_nyt(0),
		  stat_notags(0), stat_tagged(0), stat_complete(0) {}
	virtual ~StatsCollector() {}

	int get_seen() const { return stat_seen; }
	int get_onlynyt() const { return stat_onlynyt; }
	int get_nyt() const { return stat_nyt; }
	int get_notags() const { return stat_notags; }
	int get_tagged() const { return stat_tagged; }
	int get_complete() const { return stat_complete; }
};

#if 0
class PackageCounter : public PkgRecordConsumer
{
protected:
	int _count;
public:
	PackageCounter() : _count(0) {}
	virtual ~PackageCounter() {}
	virtual void operator()(pkgRecords::Parser&) { _count++; }
	int count() const throw () { return _count; }
};
#endif

template<class ITEM>
OpSet<ITEM> getItems(HierarchyNode<ITEM, Facet>* node)
{
	OpSet<ITEM> items = node->getItems();
	
	for (typename HierarchyNode<ITEM, Facet>::iterator i = node->begin();
			i != node->end(); i++)
		items += getItems(*i);

	return items;
}

// Collect items that are making a potential implication between 'existing' and
// 'missing' fail.  Basically, collect all items that have facet 'existing' but
// not facet 'missing'
template<class ITEM>
class MICollector : public Tagcoll::Consumer<ITEM, Facet>
{
protected:
	const Facet& existing;
	const Facet& missing;
	OpSet<ITEM> found;
public:
	MICollector(const Facet& existing, const Facet& missing) : existing(existing), missing(missing) {}
	virtual ~MICollector() throw() {}

	virtual void consume(const ITEM& item) throw () {}
	virtual void consume(const OpSet<ITEM>& items) throw () {}

	virtual void consume(const ITEM& item, const OpSet<Facet>& tags) throw ()
	{
		if (tags.contains(existing) && !tags.contains(missing))
			found += item;
	}

	virtual void consume(const OpSet<ITEM>& items, const OpSet<Facet>& tags) throw ()
	{
		if (tags.contains(existing) && !tags.contains(missing))
			found += items;
	}

	const OpSet<ITEM>& getFound() const throw () { return found; }
};


template<class ITEM>
OpSet<entity::Package> getMissedImplications(const Collection<ITEM, Facet>& coll,
									const Facet& existing, const Facet& missing)
{
	MICollector<ITEM> mc(existing, missing);
	coll.output(mc);
	return mc.getFound();
}

// Needs Package to get the short description
// TODO: just instantiate the package when it actually needs the short
// description
class ReportMaker
{
protected:
	component::PackageTags& debtags;
	OpSet<entity::Package> mentionedPackages;
	unsigned int itemsPerGroup;

	void listSuspects(const OpSet<entity::Package>& pkgs)
	{
		OpSet<entity::Package> toPrint = pkgs - mentionedPackages;
		if (itemsPerGroup > 0 && itemsPerGroup < toPrint.size())
		{
			float step = (float)toPrint.size() / (float)(itemsPerGroup + 1);
			int count = 0;
			float next = step;
			for (OpSet<entity::Package>::const_iterator i = toPrint.begin(); i != toPrint.end(); i++, count++)
				if (count == (int)next)
				{
					printf(" * `%.*s`_: %.*s\n", PFSTR(i->name()), PFSTR(i->shortDescription("(short description not available)")));
					mentionedPackages += *i;
					next += step;
					if (next > toPrint.size())
						break;
				}
		} else {
			for (OpSet<entity::Package>::const_iterator i = toPrint.begin(); i != toPrint.end(); i++)
				printf(" * `%.*s`_: %.*s\n", PFSTR(i->name()), PFSTR(i->shortDescription("(short description not available)")));
			mentionedPackages += toPrint;
		}
		fputs("\n", stdout);
	}

	void out(const char* text) { fputs(text, stdout); }
	void outTitle()
	{
		out(
			"Debtags TODO Report\n"
			"===================\n"
			"This is an automatically-generated report of Debian packages for which tagging seems\n"
			"to still be incomplete.  It does not aim at being a complete report, but rather it\n"
			"aims at collecting incompleteness that can be detected automatically.\n"
			"\n"
			"To make this list shorter, packages needing work are listed only once.  It might thus\n"
			"happen than when a problem is solved in a package, a further report will still list it,\n"
			"but in a different section.\n"
			"\n"
			"For more informations, please visit the Debtags website at http://debtags.alioth.debian.org\n"
			"\n"
		);
	}
	void outSpecials()
	{
		out(
			"Special-case packages\n"
			"---------------------\n"
			"Previous section: `Debtags TODO Report`_; next section: `Failed implications`_.\n"
			"\n"
			"This section lists packages whose tag set is 'strange' compared to the other packages.\n"
			"\n"
			"The list corresponds to those packages reported by ``debtags facetcoll | tagcoll findspecials``\n"
			"as being in groups with less than 100 items:\n"
			"\n"
		);

		CardinalityStore<entity::Package, Facet> coll;
		TagToFacet<entity::Package> tagStripper(coll);
		debtags.outputPatched(tagStripper);
		Facet f;
		SmartHierarchyNode<entity::Package, Facet> node(f, coll, 0);

		OpSet<Facet> seen;
		OpSet<entity::Package> specials;
		for (HierarchyNode<entity::Package, Facet>::iterator i = node.begin();
				i != node.end(); i++)
		{
			OpSet<entity::Package> items = getItems(*i);

			// Find the items in this branch that are not present in
			// any of the previous ones
			OpSet<entity::Package> newItems;
			if (!seen.empty())
			{
				for (OpSet<entity::Package>::const_iterator j = items.begin();
						j != items.end() && newItems.size() <= 101; j++)
				{
					OpSet<Tag> tags = debtags.tagdb().getTags(*j);
					bool isSeen = false;
					for (OpSet<Tag>::const_iterator i = tags.begin(); i != tags.end(); i++)
						if (seen.contains(i->facet()))
						{
							isSeen = true;
							break;
						}
					if (!isSeen && j->valid())
						newItems += *j;
				}

				if (newItems.size() <= 100)
					specials += newItems;
			}

			seen += (*i)->tag();
		}
		listSuspects(specials);
	}

	void outByExpression(const char* expr, const char* descr)
	{
		out(".. _"); out(expr); out(":\n\n");
		out(descr);
		ExprCollector<entity::Package> miss(expr);
		debtags.outputPatched(miss);
		listSuspects(miss);
	}

	void outImplications()
	{
		out(
			"Failed implications\n"
			"-------------------\n"
			"Previous section: `Special-case packages`_; next section: `Missing general tags`_.\n"
			"\n"
			"This section lists packages that have some tags A which usually go together with some tags B,\n"
			"but they don't actually any tag B.\n"
			"\n"
			"\"Why don't you fix this automatically?\", you could say.  Well, when that's about single tags,\n"
			"that can be automatically done already.  However, in this case we are looking at facets.\n"
			"For example, we are looking at packages which have uitoolkit::* tags but have no interface::*\n"
			"tags.  You can see that automatically adding the interface::* tags is an impossible task.\n"
			"\n"
		);
		outByExpression(
				"uitoolkit::* && ! interface::*",
				"Packages which have 'uitoolkit' but not 'interface' (next: `uitoolkit::* && ! implemented-in::*`_):\n\n"
		);
		outByExpression(
				"uitoolkit::* && ! implemented-in::*",
				"Packages which have 'uitoolkit' but not 'implemented-in' (previous: `uitoolkit::* && ! interface::*`_;\n"
				"next: `Missing general tags`_):\n\n"
		);
		// TODO: find more such couples
	}

	void outMissingGenerals()
	{
		out(
			"Missing general tags\n"
			"--------------------\n"
			"Previous section: `Failed Implications`; next section: `Conclusion`_.\n"
			"\n"
			"This section lists packages that are missing tags from a group that is general enough to apply\n"
			"to all packages.\n"
			"\n"
		);
		outByExpression(
				"! role::*",
				"Packages which have to ``role::`` tags (next: `Conclusion`_):\n\n"
		);
		// TODO: find more such global facets
	}
	void outConclusion()
	{
		out(
			"Conclusion\n"
			"----------\n"
			"The report ends here.  If you have ideas for more automated check, please write them\n"
			"to debtags-devel@lists.alioth.debian.org.\n"
			"\n"
			"If you are reading an HTML version of this report, don't hesitate to click on the names\n"
			"of the packages and fix the problem yourself in Erich's packagebrowser.\n"
			"\n"
			"This report has been generated by ``debtags todoreport``.\n"
		);
	}

public:
	ReportMaker(unsigned int itemsPerGroup = 0) :
		debtags(Global::get().debtags()),
		itemsPerGroup(itemsPerGroup) {}

	void printReport()
	{
		outTitle();
		outSpecials();
		outImplications();
		outMissingGenerals();

		// TODO: all packages with tags that make it assume they
		//       contain software (such as: uitoolkit), but with no
		//       implemented-in tags

		outConclusion();

		out("\n");

		// Print the links to visited packages
		for (OpSet<entity::Package>::const_iterator j = mentionedPackages.begin(); j != mentionedPackages.end(); j++)
			printf(".. _%.*s: http://debian.vitavonni.de/packagebrowser/edit.cgi?package=%.*s\n", PFSTR(j->name()), PFSTR(j->name()));
	}
};

static void printShortVocabularyItem(const Facet& facet)
{
	cout << facet.name() << " (facet) - " << facet.shortDescription() << endl;
}

static void printShortVocabularyItem(const Tag& tag)
{
	cout << tag.fullname() << " - " << tag.shortDescription() << endl;
}

static void printVocabularyItem(const Facet& tag)
{
	cout << "Facet: " << tag.name() << endl;
	cout << "Description: " << tag.shortDescription() << endl;
	cout << " " << tag.longDescription() << endl;
	cout << endl;
}
static void printVocabularyItem(const Tag& tag)
{
	cout << "Tag: " << tag.fullname() << endl;
	cout << "Description: " << tag.shortDescription() << endl;
	cout << " " << tag.longDescription() << endl;
	cout << endl;
}

#if 0
static bool matchVocabularyItem(const Tag& tag, const string& pat) throw ()
{
	if (strcasestr(tag.name().c_str(), pat.c_str())
		|| strcasestr(tag.sdesc().c_str(), pat.c_str())
		|| strcasestr(tag.ldesc().c_str(), pat.c_str()))
		return true;
	/* TODO
	else
		for (list< pair<string, string> >::const_iterator j = rec.others.begin();
				j != rec.others.end(); j++)
			if (strcasestr(j->second.c_str(), pat.c_str()))
				return true;
	*/
	return false;
/*
	if (tag.find(pat) != string::npos
		|| rec.description.find(pat) != string::npos)
		return true;
	else
		for (list< pair<string, string> >::const_iterator j = rec.others.begin();
				j != rec.others.end(); j++)
			if (j->second.find(pat) != string::npos)
				return true;
	return false;
*/
}
#endif

#if 0
class MaintCollector : public PackageConsumer
{
protected:
	DebtagsSimple<entity::Package>& debtags;
	Tagcoll::Consumer<string, Tag>& cons;
	
public:
	MaintCollector(DebtagsSimple<entity::Package>& debtags, Tagcoll::Consumer<string, Tag>& cons)
		throw () : debtags(debtags), cons(cons) {}
	void consume(const entity::Package& pkg) throw ()
	{
		cons.consume(pkg.maint(), debtags.tagDB().getTags(pkg));
	}
};
#endif

class MaintPrinter : public Tagcoll::Consumer<string, Tag>
{
protected:
	void consumeItemUntagged(const string& m) throw ()
	{
		printf("%.*s\n", PFSTR(m));
	}
	void consumeItem(const string& m, const OpSet<Tag>& tags) throw ()
	{
		printf("%.*s: ", PFSTR(m));
		for (OpSet<Tag>::const_iterator j = tags.begin();
				j != tags.end(); j++)
			if (j != tags.begin())
				printf(", %.*s", PFSTR(j->fullname()));
			else
				printf("%.*s", PFSTR(j->fullname()));
		printf("\n");
	}
public:
};

class CommandlineParserWithCommand : public CommandlineParser
{
protected:
	map<string, int> command_map;

public:
	CommandlineParserWithCommand(const std::string& argv0,
									const std::string& cmdline_summary,
									const std::string& description) throw ()
		: CommandlineParser(argv0, cmdline_summary, description)
	{
		add("version", 'V', "version", "print the program version, then exit");
	}

	void addCommand(const std::string name, int id) throw ()
	{
		command_map[name] = id;
	}

	int parse(int& argc, const char**& argv) throw ()
	{
		if (!CommandlineParser::parse(argc, argv))
		{
			printHelp();
			exit(1);
		}
		if (get("help").defined())
		{
			printHelp();
			exit(0);
		}
		if (get("version").defined())
		{
			printf("%s ver." PACKAGE_VERSION "\n", APPNAME);
			exit(0);
		}
		if (argc == 1)
		{
			printHelp();
			exit(1);
		}

		string command_string = argv[1];
		map<string, int>::const_iterator cmap_i = command_map.find(command_string);
		if (cmap_i == command_map.end())
		{
			fprintf(stderr, "Invalid command: \"%.*s\"\n", PFSTR(command_string));
			printHelp();
			exit(1);
		}

		for (int i = 1; i < argc; i++)
			argv[i] = argv[i + 1];
		--argc;
		
		return cmap_i->second;
	}
};

class CommandlineArgs
{
protected:
	int argc;
	const char** argv;
	int _next;

public:
	CommandlineArgs(int argc, const char* argv[]) throw () : argc(argc), argv(argv), _next(1) {}

	// Return true if there is another argument left in the list
	bool hasNext() const throw () { return argc >= _next + 1; }

	// Return the next argument in the list
	string next() throw ()
	{
		if (hasNext())
		{
			return argv[_next++];
		} else {
			return "-";
		}
	}
};

enum valid_command { UPDATE, CHECK, TAGSHOW, TAGSEARCH, TAGCAT, SHOW, RELATED, CAT, SEARCH, GREP, INSTALL, MKPATCH, MAINTAINERS, TAG, SUBMIT, TODO, SCORE, FACETCOLL, STATS, TODOREPORT, SMARTSEARCH };

int main(int argc, const char* argv[])
{
	try {
		// Install the handler for unexpected exceptions
		InstallUnexpected installUnexpected;

		CommandlineParserWithCommand opts(APPNAME, "[options] [command] [file1|-] [file2|-]",
				"Perform various operations on a tagged collection\n\n"
				"Commands are:\n"
				"  update        Updates the package tag database (requires root)\n"
				"  check <file>	 Check that all the tags in the given tagged collection are present\n"
				"                in the tag vocabulary.  Checks the main database if no file is\n"
				"                specified\n"
				"  tagcat <tag>	 Output the tag vocabulary\n"
				"  tagshow <tag> Show the vocabulary informations about a tag\n"
				"  tagsearch <string [string [string [...]]]>\n"
				"                Show a summary of all tags whose data contains the given strings\n"
				"  show <pkg>    Call apt-cache show <pkg>, but add tag informations to the output.\n"
				"  related <pkg1[,pkg2[,pkg2,[...]]]>\n"
				"                Show packages related to the specified ones.\n"
				"  cat           Output the full package tag database\n"
				"  search [-v] <tag expression>\n"
				"                Output the names and description of the packages that match\n"
				"                the given tag expression\n"
				"  grep [-v] [-q] <tag expression>\n"
				"                Output the lines of the full package tag database that match\n"
				"                the given tag expression\n"
				"  install [-v] [-q] <tag expression>\n"
				"                apt-get install the packages that match the given tag expression\n"
				"  mkpatch [filename]\n"
				"                Create a tag patch between the current tag database and the tag\n"
				"                collection [filename]\n"
				"  maintainers   Create a tagged collection of maintainers and the tags of the\n"
				"                packages they maintain\n"
				"  tag [add <package> <tags...>\n"
				"  tag [rm  <package> <tags...>\n"
				"  tag [ls  <package>\n"
				"                View and edit the tags for a package\n"
				"  submit [patch]\n"
				"                Mail the given patch file to the central tag repository.\n"
				"                If [patch] is omitted, mail the local tag modifications.\n"
				"  todo          Print a list of the installed packages that are not yet tagged\n"
				"  score         Score uninstalled packages according to how often their tags\n"
				"                appear in the packages that are installed already\n"
				"  facetcoll     Print the tagged collection where each package is tagged with\n"
				"                its facets only\n"
				"  stats         Print statistics about Debtags\n"
				"  todoreport    Print a report of packages needing work\n"
				"  ssearch <word [word1 [+tag [-tag1 ...]]]>\n"
				"                Perform a keyword search integrated with related packages.\n"
				"                A + prefix indicates a wanted tag.  A - prefix indicates\n"
				"                an unwanted tag.  Other words indicate keywords to search.\n"
				"                Remember to use '--' before unwanted tags to avoid to have\n"
				"                them interpreted as commandline switches.\n");


		/*
		opts.add("hierarchy", 's', "smart-hierarchy", "build a smart hierarchy");
		opts.add("implications", 'm', "show-implications", "output a list of tag implications");
		opts.add("copy", 'c', "copy", "output the collection");
		opts.add("diff", 'd', "diff", "output a tag patch file with the differences between two files (requires two file arguments)");
		*/

		//opts.add("expanded", 'x', "expanded-output", "produce full (and redundant) output data instead of compact");
		opts.add("groupitems", 'g', "group-items", "group items with the same tagset in the output collection");
		opts.add("distance", 'd', "distance", "set the maximum distance to use for the \"related\" command (defaults to 0)", "num");

		opts.add("invert-match", 'v', "invert-match", "invert the sense of matching, to select non-matching lines");
		opts.add("quiet", 'q', "quiet", "do not write anything to standard output, but exit with 0 if any match is found");
		
		opts.add("verbose", 'V', "verbose", "enable verbose output");
		opts.add("debug", 0, "debug", "enable debugging output (including verbose output)");

		opts.addCommand("update", (int)UPDATE);
		opts.addCommand("check", (int)CHECK);
		opts.addCommand("tagcat", (int)TAGCAT);
		opts.addCommand("tagshow", (int)TAGSHOW);
		opts.addCommand("tagsearch", (int)TAGSEARCH);
		opts.addCommand("show", (int)SHOW);
		opts.addCommand("related", (int)RELATED);
		opts.addCommand("cat", (int)CAT);
		opts.addCommand("search", (int)SEARCH);
		opts.addCommand("grep", (int)GREP);
		opts.addCommand("install", (int)INSTALL);
		opts.addCommand("mkpatch", (int)MKPATCH);
		opts.addCommand("maintainers", (int)MAINTAINERS);
		opts.addCommand("tag", (int)TAG);
		opts.addCommand("submit", (int)SUBMIT);
		opts.addCommand("todo", (int)TODO);
		opts.addCommand("score", (int)SCORE);
		opts.addCommand("facetcoll", (int)FACETCOLL);
		opts.addCommand("stats", (int)STATS);
		opts.addCommand("todoreport", (int)TODOREPORT);
		opts.addCommand("ssearch", (int)SMARTSEARCH);

		// Process the commandline
		valid_command cmd = (valid_command)opts.parse(argc, argv);

		CommandlineArgs args(argc, argv);

		if (opts.get("verbose").defined())
			::Environment::get().verbose(true);

		if (opts.get("debug").defined())
			::Environment::get().debug(true);

		/*
		if (cmd != UPDATE && cmd != CAT && cmd != GREP && cmd != INSTALL && cmd != TAGSHOW)
			Debtags::Environment::init(false);
		*/
		
		// Perform the correct operation
		switch (cmd)
		{
			// Output the full package tag database
			case CAT:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();
				auto_ptr< Printer<Package, Tag> > printer;

				if (opts.get("groupitems").defined())
					printer = auto_ptr< Printer<Package, Tag> >(new GroupedTagcollPrinter);
				else
					printer = auto_ptr< Printer<Package, Tag> >(new TagcollPrinter);

				Searcher searcher(debtags, printer.get());
				searcher.output();

				return 0;
			}

			// search [-v] <tag expression>\n"
			// Output the names and description of the packages that match\n"
			// the given tag expression\n"
			case SEARCH:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();
				APTPrinter printer;
				Searcher searcher(debtags, &printer);
				string expression = args.next();

				int matched =
					searcher.output(expression, opts.get("invert-match").defined());

				return matched > 0 ? 0 : 1;
			}
			

			// grep [-v] [-q] <tag expression>
			// Output the lines of the full package tag database that match the
			// given tag expression
			case GREP:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();
				auto_ptr< Printer<Package, Tag> > printer;

				if (opts.get("quiet").defined())
					printer = auto_ptr< Printer<Package, Tag> >(new NullPrinter);
				if (opts.get("groupitems").defined())
					printer = auto_ptr< Printer<Package, Tag> >(new GroupedTagcollPrinter);
				else
					printer = auto_ptr< Printer<Package, Tag> >(new TagcollPrinter);

				Searcher searcher(debtags, printer.get());
				string expression = args.next();

				int matched =
					searcher.output(expression, opts.get("invert-match").defined());

				return matched > 0 ? 0 : 1;
			}

			// install [-v] [-q] <tag expression>
			// apt-get install the packages that match the given tag expression
			case INSTALL:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();
				Installer installer;

				Searcher searcher(debtags, &installer);
				string expression = args.next();

				searcher.output(expression, opts.get("invert-match").defined());

				return 1;
			}

			// tagcat
			// Output the entire tag vocabulary
			case TAGCAT:
			{
				debtagsInit();
				wantTagDatabase();

				component::Tags& voc = Global::get().tags();

				OpSet<Facet> facets = voc.facets();
				for (OpSet<Facet>::const_iterator i = facets.begin();
						i != facets.end(); i++)
				{
					printVocabularyItem(*i);

					OpSet<Tag> tags = i->tags();
					for (OpSet<Tag>::const_iterator j = tags.begin();
							j != tags.end(); j++)
						printVocabularyItem(*j);
				}
				return 0;
			}
			// tagshow <tag>
			// Show the vocabulary informations about a tag
			case TAGSHOW:
			{
				debtagsInit();
				wantTagDatabase();
				string tag = args.next();

				Tag t = Global::get().tags().tagByName(tag);
				if (!t)
				{
					verbose("Tag `%.*s' was not found in tag vocabulary\n", PFSTR(tag));
					return 1;
				}
				else
				{
					printVocabularyItem(t);
					return 0;
				}
			}
			// tagsearch <pattern [pattern [pattern [...]]]>
			// Show a summary of all tags matching the given patterns
			case TAGSEARCH:
			{
				debtagsInit();
				wantTagDatabase();

				SubstringTagMatcher match;
				
				// Get the patterns to be matched
				bool empty;
				while (args.hasNext())
				{
					string pattern = args.next();
					match.add(pattern);
					empty = false;
				}

				if (empty)
				{
					error("No patterns given in commandline\n");
					return 1;
				}

				component::Tags& voc = Global::get().tags();

				int matched = 0;

				OpSet<Facet> facets = voc.facets();
				for (OpSet<Facet>::const_iterator i = facets.begin();
						i != facets.end(); i++)
				{
					if (match(*i))
					{
						matched++;
						printShortVocabularyItem(*i);
					}

					OpSet<Tag> tags = i->tags();
					for (OpSet<Tag>::const_iterator j = tags.begin();
							j != tags.end(); j++)
						if (match(*j))
						{
							matched++;
							printShortVocabularyItem(*j);
						}
				}

				return matched > 0 ? 0 : 1;
			}
			// show <pkg>
			// Call apt-cache show <pkg>, but add tag informations to the output.\n"
			case SHOW:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				while (args.hasNext())
				{
					string name = args.next();
					
					entity::Package pkg = packageByName(name);
					if (pkg.valid())
					{
						cout << pkg.candidateVersion().completeRecord() << endl;
						cout << "Tags: ";
						OpSet<Tag> ts = debtags.tagdb().getTags(pkg);
						for (OpSet<Tag>::const_iterator i = ts.begin();
								i != ts.end(); i++)
							if (i == ts.begin())
								cout << i->fullname();
							else
								cout << ", " + i->fullname();
						cout << endl;
						return 0;
					} else {
						verbose("Package %.*s not found", PFSTR(name));
						return 1;
					}
				}
			}
			// related <pkg1[,pkg2[,pkg2,[...]]]>
			// Show packages related to the specified ones
			case RELATED:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				int maxdist = 0;
				if (opts.get("distance").defined())
					maxdist = opts.get("distance").intVal();
				string pkg = args.next();

				// Split the items on commas
				string splititem;
				OpSet<entity::Package> splititems;
				for (string::const_iterator c = pkg.begin(); c != pkg.end(); c++)
					if (*c == ',')
					{
						entity::Package p = packageByName(splititem);
						if (p.valid())
						{
							splititems.insert(p);
						} else {
							error("Item \"%.*s\" does not exist in the collection\n", PFSTR(splititem));
							return 1;
						}
						splititem = string();
					} else
						splititem += *c;
				entity::Package p = packageByName(splititem);
				if (p.valid())
				{
					splititems.insert(p);
				} else {
					error("Item \"%.*s\" does not exist in the collection\n", PFSTR(splititem));
					return 1;
				}

				// Get the tagset as the intersection of the tagsets of all input items
				OpSet<entity::Package>::const_iterator i = splititems.begin();
				OpSet<Tag> ts = debtags.tagdb().getTags(*i);
				for (++i; i != splititems.end(); i++)
					ts = ts ^ debtags.tagdb().getTags(*i);

				if (ts.empty())
				{
					if (splititems.size() > 1)
						fprintf(stderr, "The packages %.*s are unrelated: cannot find a barycenter to start computing relationships from.\n", PFSTR(pkg));
					else
						fprintf(stderr, "The package %.*s has no tags attached.\n", PFSTR(pkg));
					return 1;
				}

				APTPrinter printer(splititems);
				OpSet<entity::Package> related(debtags.tagdb().getRelatedItems(ts, maxdist));
				printer.print(related);
				printer.flush();

				if (printer.count() > 50 && maxdist == 0 && isatty(1))
				{
					string tags;
					for (OpSet<Tag>::const_iterator i = ts.begin();
							i != ts.end(); i++)
						if (i == ts.begin())
							tags += i->fullname();
						else
							tags += "%2C" + i->fullname();
					feedback("\nIt seems that this set of packages lacks tag information that could help to better distinguish package similarities.\nYou can help providing better tagging: just point your browser to http://debian.vitavonni.de/packagebrowser/?tags=%.*s\n", PFSTR(tags));
				}
				return 0;
			}
			// check <file>
			// Check that all the tags in the given tagged collection are
			// present in the tag vocabulary.  Checks the main database if no
			// file is specified
			case CHECK:
			{
				debtagsInit();
				wantTagDatabase();

				string file;
				if (args.hasNext())
					file = args.next();
				else
					file = utils::Path::tagdb();

				TagcollChecker checker;
				readCollection(file, checker);

				if (checker.missingCount() > 0)
				{
					checker.report();
					return 1;
				}
				else
					return 0;
			}
			// update
			// Updates the package tag database (requires root)
			case UPDATE:
			{
				if (geteuid() != 0)
				{
					throw ConsistencyCheckException("You must be root to update the system debtags database");
				}
				struct winsize ws;
				unsigned int ScreenWidth;

				if (ioctl(1, TIOCGWINSZ, &ws) != -1 && ws.ws_col >= 5)
					ScreenWidth = ws.ws_col - 1;

				AcqTextStatus Stat(ScreenWidth,_config->FindI("quiet",0));
				//Debtags::Environment::get().updateDebtagsDatabase(&Stat);
				mode_t prev_umask = umask(022);
				debtags::updateDatabase(&Stat);
				umask(prev_umask);
				feedback("Done.\n");
				break;
			}
			// mkpatch [filename]
			// Create a tag patch between the current tag database and the tag
			// collection [filename]
			case MKPATCH:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				string file = args.next();

				InputMerger<entity::Package, Tag> coll;
				Converter<string, Package> toitem;
				Converter<string, Tag> totag;
				readCollection(toitem, totag, file, coll);

				PatchList<entity::Package, Tag> newpatches;
				newpatches.addPatch(debtags.tagdb(), coll);

				Converter<entity::Package, string> fromitem;
				Converter<Tag, string> fromtag;
				TextFormat<entity::Package, Tag>::outputPatch(fromitem, fromtag, newpatches, stdout);
				break;
			}
			// maintainers
			// Create a tagged collection of maintainers and the tags of the
			// packages they maintain
			case MAINTAINERS:
			{
				debtagsInit();
				wantTagDatabase();

				// Gather maintainer informations
				InputMerger<string, Tag> maints;
				for (component::Packages::iterator i = Global::get().packages().packagesBegin();
						i != Global::get().packages().packagesEnd(); i++)
				{
					if (!i->valid())
						continue;
					entity::Version v = i->candidateVersion();
					maints.consume(v.maintainer(), i->tags());
				}
				MaintPrinter printer;
				maints.output(printer);
				break;
			}
			// tag
			//   tag [add <package> <tags...>\n"
			//   tag [rm  <package> <tags...>\n"
			//   tag [ls  <package>\n"
			//                View and edit the tags for a package\n");
			case TAG:
			{
				std::string cmd = args.next();

				if (cmd == "add" || cmd == "rm")
				{
					component::PackageTags& debtags = debtagsInit(true);
					wantTagDatabase();

					string name = args.next();

					entity::Package pkg = packageByName(name);
					if (!pkg.valid())
					{
						error("Package %.*s not found\n", PFSTR(name));
						return 1;
					}

					OpSet<Tag> tagset;

					while (args.hasNext())
					{
						string tag = args.next();
						Tag t = Global::get().tags().tagByName(tag);
						if (t)
							tagset += t;
						else
							error("Tag `%.*s' not found: ignored\n", PFSTR(tag));
					}
					
					if (!tagset.empty())
					{
						PatchList<entity::Package, Tag> change;
						if (cmd == "add")
							change.addPatch(Patch<entity::Package, Tag>(pkg, tagset, OpSet<Tag>()));
						else
							change.addPatch(Patch<entity::Package, Tag>(pkg, OpSet<Tag>(), tagset));
						debtags.tagdb().applyChange(change);
						debtags.savePatch();
					} else
						verbose("No tags to add\n");
				}
				else if (cmd == "ls")
				{
					component::PackageTags& debtags = debtagsInit();
					wantTagDatabase();

					string name = args.next();
					entity::Package pkg = packageByName(name);
					if (pkg.valid())
					{
						OpSet<Tag> ts = debtags.tagdb().getTags(pkg);
						for (OpSet<Tag>::const_iterator i = ts.begin();
								i != ts.end(); i++)
							printf("%.*s\n", PFSTR(i->fullname()));
						return 0;
					} else {
						verbose("Package %.*s not found", PFSTR(name));
						return 1;
					}
				}
				else
					throw ConsistencyCheckException("command " + cmd + " is not valid working with tags");
				break;
			}
			// submit
			// Mail the local updates to the tag database to the central tag
			// repository
			case SUBMIT:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				if (args.hasNext())
				{
					StdioParserInput in(args.next());
					Converter<string, Package> toitem;
					Converter<string, Tag> totag;
					PatchList<entity::Package, Tag> patch =
							TextFormat<entity::Package, Tag>::parsePatch(toitem, totag, in);
					debtags.sendPatch(patch);
				}
				else
					debtags.sendPatch();

				break;
			}
			// todo
			// Print a list of the installed packages that are not yet tagged
			case TODO:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				// Write the package names to stdout
				APTPrinter printer;

				// Filter to select the right packages
				TODOFilter filter(printer);

				debtags.outputPatched(filter);

				printer.flush();

				break;
			}
			// score
			// Score uninstalled packages according to how often their tags
			// appear in the packages that are installed already
			case SCORE:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				// Compute tag scores
				TagsScorer tscorer;
				debtags.outputPatched(tscorer);

				// Compute package scores
				PackageScorer pscorer(tscorer);
				debtags.outputPatched(pscorer);

				// Print the results
				for (PackageScorer::const_iterator i = pscorer.begin();
						i != pscorer.end(); i++)
					printf("%d %.*s - %.*s\n", i->first,
							PFSTR(i->second.name()), PFSTR(i->second.shortDescription("(short description not available)")));
				break;
			}
			// facetcoll
			// Print the tagged collection where each package is tagged with
			// its facets only
			case FACETCOLL:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				FacetcollPrinter<entity::Package> printer;
				TagToFacet<entity::Package> tagToFacet(printer);
				debtags.outputPatched(tagToFacet);
				break;
			}
			// stats
			// Print statistics about Debtags
			case STATS:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();

				int pkgCount = Global::get().packages().packageCount();
				printf("Total count of packages: %d\n", pkgCount);

				StatsCollector stats;
				debtags.outputPatched(stats);

				printf("Total count of packages (according to APT): %d\n", pkgCount);
				printf("Total count of packages (according to Debtags): %d\n", stats.get_seen());

				const component::Tags& voc = Global::get().tags();
				printf("Number of facets: %d\n", voc.facets().size());
				printf("Number of tags: %d\n", voc.tags().size());

				// Copied from Debtags class: compute the toplevel facets
				// TODO: use Debtags instead of Environment throughout all Debtags
				CardinalityStore<entity::Package, Facet> coll;
				TagToFacet<entity::Package> tagStripper(coll);
				debtags.outputPatched(tagStripper);
				Facet f;
				SmartHierarchyNode<entity::Package, Facet> node(f, coll, 0);
				printf("Number of automatically computed toplevel facets: %d\n", node.size());

				printf("Number of packages with special::completely-tagged tags: %d (%.1f%%)\n",
						stats.get_complete(), (float)stats.get_complete()*100/stats.get_seen());
				printf("Number of packages with tags, but no special::not-yet-tagged tags: %d (%.1f%%)\n",
						stats.get_tagged(), (float)stats.get_tagged()*100/stats.get_seen());
				printf("Number of packages with special::not-yet-tagged tags: %d (%.1f%%)\n",
						stats.get_nyt(), (float)stats.get_nyt()*100/stats.get_seen());
				printf("Number of packages with only special::not-yet-tagged tags: %d (%.1f%%)\n",
						stats.get_onlynyt(), (float)stats.get_onlynyt()*100/stats.get_seen());
				printf("Number of packages with no tags: %d (%.1f%%)\n",
						stats.get_notags(), (float)stats.get_notags()*100/stats.get_seen());

				break;
			}
			// todoreport
			// Print a report of packages needing work
			case TODOREPORT:
			{
				debtagsInit();
				wantTagDatabase();

				unsigned int itemsPerGroup = 0;
				if (args.hasNext())
					itemsPerGroup = atoi(args.next().c_str());

				ReportMaker rm(itemsPerGroup);
				rm.printReport();

				break;
			}
			// ssearch <word [word1 [word2 ...]]>
			// Perform a keyword search integrated with related packages
			case SMARTSEARCH:
			{
				component::PackageTags& debtags = debtagsInit();
				wantTagDatabase();
				APTPrinter printer;
				SmartSearcher smart(debtags, &printer);
				Searcher searcher(debtags, &smart);
				predicate::Predicate<Package> p = predicate::True<Package>();
					//predicate::Factory<Package>::description(args.next());
				while (args.hasNext())
				{
					string arg = args.next();
					switch (arg[0])
					{
						case '+':
							p = p and predicate::Factory<Package>::tag(
									Global::get().tags().tagByName(arg.substr(1)));
							break;
						case '-':
							p = p and not predicate::Factory<Package>::tag(
									Global::get().tags().tagByName(arg.substr(1)));
							break;
						default:
							p = p and predicate::Factory<Package>::description(arg);
							break;
					}
				}

				int step1 = searcher.output(p);
				int step2 = smart.outputRelated();

				cout << step1 << " normal matches plus " << step2 << " related packages." << endl;
				OpSet<Tag> topTags = smart.topTags();
				cout << "Top tags were: " << topTags << endl;

				return step1 + step2 > 0 ? 0 : 1;
			}
			

		}

		return 0;
	}
	catch (Exception& e)
	{
		fatal_error("%s: %.*s\n", e.type(), PFSTR(e.desc()));
		return 1;
	}
}

// vim:set ts=4 sw=4:
